/*
frequency_visualizer.c
Created by Heagan Ahmed (hsa35), Corey Chang (clc282), Isaac Kruger (idk4)
for ECE4760 Fall 2014 Final Project

Uses implementations from the program dct1d_test.c
created mer jan  2 13:55:12 GMT 2002
by Yann Guidon <whygee@f-cpu.org> and
modified for GCC/mega644 by Bruce Land, BRL4@cornell.edu
and extended to 16 and 32 samples
Feb 2009
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include "uart.h"
#include <math.h>

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
       

#define begin {
#define end }

#define DISP_ALL 0			// Display all 32 coefficients (8 sets of 4)
#define DISP_LOWER_16 1		// Display only lower 16 coefficients
#define DISP_UPPER_16 2		// Display only upper 16 coefficients
#define DISP_ODD 3			// Display only odd coefficients
#define DISP_EVEN 4			// Display only even coefficients

//State machine state names
#define NoPush 1 
#define MaybePush 2
#define Pushed 3
#define MaybeNoPush 4

#define int2fix(a)   (((int)(a))<<8) 
#define float2fix(a) ((int)((a)*256.0)) 
#define fix2float(a) ((float)(a)/256.0)  

#define t1 1

// Fast fixed point multiply assembler macro
#define multfix(a,b)          	  \
({                                \
int prod, val1=a, val2=b ;        \
__asm__ __volatile__ (            \ 
"muls %B1, %B2	\n\t"              \
"mov %B0, r0 \n\t"	               \ 
"mul %A1, %A2\n\t"	               \ 
"mov %A0, r1   \n\t"              \ 
"mulsu %B1, %A2	\n\t"          \ 
"add %A0, r0  \n\t"               \ 
"adc %B0, r1 \n\t"                \ 
"mulsu %B2, %A1	\n\t"          \ 
"add %A0, r0 \n\t"           \ 
"adc %B0, r1  \n\t"          \ 
"clr r1  \n\t" 		         \ 
: "=&d" (prod)               \
: "a" (val1), "a" (val2)      \
);                            \
prod;                        \
})


#define pi 3.14159


int  m1, m2, m3, m4 ;   /* 8 sample twiddle factors -- needed for 32 points */
int c1,c3,c5,c7,c9,c11,c13,c15; //16 sample factors -- needed for 32 points   
int c17,c19,c21,c23,c25,c27,c29,c31; //32 sample factors
int a[8];  /* the 8 sample inputs -- needed for 32 points*/
int S[8];  /* the 8 sample outputs -- needed for 32 points */  
int a16[16];  /* the 16 sample inputs  -- needed for 32 points*/
int S16[16];  /* the 16 sample outputs  -- needed for 32 points*/
int a32[32];  /* the 16 sample inputs */
int S32[32];  /* the 16 sample outputs */
int a4[4];
int S4[4];
int a2[2];
int S2[2];
int bufferIn[32];;

volatile char buffer_i;
volatile int overflow_ct;
volatile char proc_flag;
volatile char sample_count;

volatile char state;
volatile char next_state;
volatile int debounce_time;
volatile unsigned char PushState;
volatile char print_flag;

ISR( TIMER0_OVF_vect )
{
	// Load the ADC value into the buffer
	bufferIn[buffer_i] = ADCH;
	buffer_i++;

	// If the buffer is full, then we are ready to process the sample
	if (buffer_i == 32){
		buffer_i = 0;
		proc_flag = 1;
	}

	// Start a new ADC conversion
	ADCSRA |= (1<<ADSC);
	
	// Debounce any button push
	if(debounce_time > 0)
	{
		debounce_time--;
	}

	else if(debounce_time == 0)
	{
		debounce_time=t1;     //reset the task timer
		switch (PushState)
		begin
		case NoPush: 
			if (~PINB & 0x10) PushState=MaybePush;
			else PushState=NoPush;
			break;
		case MaybePush:
			if (~PINB & 0x10) 
			begin
			next_state = (state+1)%5;
			print_flag = 1;
			PushState=Pushed;   
			end
			else PushState=NoPush;
			break;
		case Pushed:  
			if (~PINB & 0x10)PushState=Pushed; 
			else PushState=MaybeNoPush;    
			break;
		case MaybeNoPush:
			if (~PINB & 0x10) PushState=Pushed; 
			else 
			begin
			PushState=NoPush;
			end    
			break;
		end
	}
}

//==================================================================  
void DCT1D_02(void)
begin


/*Modified for GCC by Bruce Land BRL4@cornell.edu
DCT1D_02.c
version mar jan  1 08:45:57 GMT 2002
whygee@f-cpu.org

1* 8-bin DCT for a "baseline" JPG compressor.

originally cut and pasted from : sbcci_DCT2D.pdf
"Pipelined Fast 2-D DCT Architecture for JPEG Image Compression"
Luciano Volcan Agostini  <agostini@inf.ufrgs.br>
Ivan Saraiva Silva       <ivan@dimap.ufrn.br>
Sergio Bampi             <bampi@inf.ufrgs.br>*/

  
	int b0, b1, b2, b3, b4, b5, b6, b7;
    int c0, c1, c2, c3, c4, c5, c6;
    int d0, d1, d3, d4;
    int e2, e3, e4, e6, e7;
    int f2, f3, f4, f5, f6, f7;

    /* Step 1 */
    b0 = a[0] + a[7];
    b1 = a[1] + a[6];
    b2 = a[3] - a[4]; /* corrected */
    b3 = a[1] - a[6];
    b4 = a[2] + a[5];
    b5 = a[3] + a[4];
    b6 = a[2] - a[5];
    b7 = a[0] - a[7];  /***/

    
    /* Step 2 */
    c0 = b0 + b5;
    c1 = b1 - b4;
    c2 = b2 + b6;  /***/
    c3 = b1 + b4;
    c4 = b0 - b5;
    c5 = b3 + b7;  /***/
    c6 = b3 + b6;  /***/
    
    /* Step 3 */
    d0 = c0 + c3;  /***/
    d1 = c0 - c3;  /***/
    d3 = c1 + c4;
    d4 = c2 - c5; 
    
    /* Step 4 */
    e2 = multfix(m3,c2);  /*c2*/
    e3 = multfix(m1,c6);  /*c6*/
    e4 = multfix(m4,c5);  /*c5*/
    e6 = multfix(m1,d3);
    e7 = multfix(m2,d4);
    
    /* Step 5 */
    f2 = c4 + e6;  /*c4*/
    f3 = c4 - e6;  /*c4*/
    f4 = e3 + b7;  /*b7*/
    f5 = b7 - e3;  /*b7*/
    f6 = e2 + e7;
    f7 = e4 + e7;
    
    /* Step 6 */
    S[0] = d0;       /*d0*/
    S[1] = f4 + f7;
    S[2] = f2;
    S[3] = f5 - f6;
    S[4] = d1;       /*d1*/
    S[5] = f5 + f6;
    S[6] = f3;
    S[7] = f4 - f7;

end


//=========================================================  
// From the algorithm to combine 2 8-point DCTs given in
// KR Rao and P Yip:
// "Discrete Cosine Transform", Academic Press 1990
// pages 60-61
void DCT1D_16(void)
begin
    //segment 16 samples into first 8 
    //then call DCT1D_02 
    //and assign outputs to correct frequencies
    a[0] = a16[0] + a16[15];
    a[1] = a16[1] + a16[14];
    a[2] = a16[2] + a16[13];
    a[3] = a16[3] + a16[12];
    a[4] = a16[4] + a16[11];
    a[5] = a16[5] + a16[10];
    a[6] = a16[6] + a16[9];
    a[7] = a16[7] + a16[8];
    DCT1D_02(); 
    S16[0] = S[0];
    S16[2] = S[1];
    S16[4] = S[2];
    S16[6] = S[3]; 
    S16[8] = S[4];
    S16[10] = S[5];
    S16[12] = S[6];
    S16[14] = S[7];
    
    //segment 16 samples into second 8 
    //premult by factors 
    //then call DCT1D_02 
    //and assign outputs to correct frequencies
    a[0] = multfix(c1, (a16[0] - a16[15]));
    a[1] = multfix(c3, (a16[1] - a16[14]));
    a[2] = multfix(c5, (a16[2] - a16[13]));
    a[3] = multfix(c7, (a16[3] - a16[12])); 
    a[4] = multfix(c9, (a16[4] - a16[11]));
    a[5] = multfix(c11, (a16[5] - a16[10]));
    a[6] = multfix(c13, (a16[6] - a16[9]));
    a[7] = multfix(c15, (a16[7] - a16[8]));
    DCT1D_02(); 
    S16[1] = S[0] + S[1];
    S16[3] = S[1] + S[2];
    S16[5] = S[2] + S[3];
    S16[7] = S[3] + S[4]; 
    S16[9] = S[4] + S[5];
    S16[11] = S[5] + S[6];
    S16[13] = S[6] + S[7];
    S16[15] = S[7] ;
  
end

//=========================================================  
// From the algorithm to combine 2 16-point DCTs given in
// KR Rao and P Yip:
// "Discrete Cosine Transform", Academic Press 1990
// pages 60-61
void DCT1D_32(void)
begin
    //segment 16 samples into first 16 
    //then call DCT1D_16 
    //and assign outputs to correct frequencies
    a16[0] = a32[0] + a32[31];
    a16[1] = a32[1] + a32[30];
    a16[2] = a32[2] + a32[29];
    a16[3] = a32[3] + a32[28];
    a16[4] = a32[4] + a32[27];
    a16[5] = a32[5] + a32[26];
    a16[6] = a32[6] + a32[25];
    a16[7] = a32[7] + a32[24];
    a16[8] = a32[8] + a32[23];
    a16[9] = a32[9] + a32[22];
    a16[10] = a32[10] + a32[21];
    a16[11] = a32[11] + a32[20];
    a16[12] = a32[12] + a32[19];
    a16[13] = a32[13] + a32[18];
    a16[14] = a32[14] + a32[17];
    a16[15] = a32[15] + a32[16];
    DCT1D_16(); 
    S32[0] = S16[0];
    S32[2] = S16[1];
    S32[4] = S16[2];
    S32[6] = S16[3]; 
    S32[8] = S16[4];
    S32[10] = S16[5];
    S32[12] = S16[6];
    S32[14] = S16[7];
    S32[16] = S16[8];
    S32[18] = S16[9];
    S32[20] = S16[10];
    S32[22] = S16[11]; 
    S32[24] = S16[12];
    S32[26] = S16[13];
    S32[28] = S16[14];
    S32[30] = S16[15];
    
    //segment 16 samples into second 16 
    //premult by factors 
    //then call DCT1D_16
    //and assign outputs to correct frequencies
    a16[0] = multfix(c1, (a32[0] - a32[31]));
    a16[1] = multfix(c3, (a32[1] - a32[30]));
    a16[2] = multfix(c5, (a32[2] - a32[29]));
    a16[3] = multfix(c7, (a32[3] - a32[28])); 
    a16[4] = multfix(c9, (a32[4] - a32[27]));
    a16[5] = multfix(c11, (a32[5] - a32[26]));
    a16[6] = multfix(c13, (a32[6] - a32[25]));
    a16[7] = multfix(c15, (a32[7] - a32[24]));
    a16[8] = multfix(c17, (a32[8] - a32[23]));
    a16[9] = multfix(c19, (a32[9] - a32[22]));
    a16[10] = multfix(c21, (a32[10] - a32[21]));
    a16[11] = multfix(c23, (a32[11] - a32[20])); 
    a16[12] = multfix(c25, (a32[12] - a32[19]));
    a16[13] = multfix(c27, (a32[13] - a32[18]));
    a16[14] = multfix(c29, (a32[14] - a32[17]));
    a16[15] = multfix(c31, (a32[15] - a32[16]));
    DCT1D_16(); 
    S32[1] = S16[0] + S16[1];
    S32[3] = S16[1] + S16[2];
    S32[5] = S16[2] + S16[3];
    S32[7] = S16[3] + S16[4]; 
    S32[9] = S16[4] + S16[5];
    S32[11] = S16[5] + S16[6];
    S32[13] = S16[6] + S16[7];
    S32[15] = S16[7] + S16[8];
    S32[17] = S16[8] + S16[9];
    S32[19] = S16[9] + S16[10];
    S32[21] = S16[10] + S16[11];
    S32[23] = S16[11] + S16[12]; 
    S32[25] = S16[12] + S16[13];
    S32[27] = S16[13] + S16[14];
    S32[29] = S16[14] + S16[15];
    S32[31] = S16[15] ;
end


// Copies contents of array1 into array2
void arrayCopy(int array1[32], int array2[32])
{
	for (int i = 0; i<32; i++) {
		array2[i] = array1[i];
	}
}


// Normalizes the input to be between -1 and 1 in fixed pt
// for use in the DCT
void normalizeForDCT()
{
	for(int i=0; i<32; i++)
	{
		a32[i] = float2fix(((float)a32[i])/128-1);
	}
}


// Turns on the LEDs for a given frequency group
void turn_on_LED(char c, int i)
{
	if(i ==0 ){
		if (0x01 & c)
		{
			PORTB |= (1<<PORTB0);
		}
		if (0x02 & c)
		{
			PORTB |= (1<<PORTB1);
		}
		if (0x04 & c)
		{
			PORTB |= (1<<PORTB2);
		}
	}
	else if(i == 1){
		if (0x01 & c)
		{
			PORTA |= (1<<PORTA1);
		}
		if (0x02 & c)
		{
			PORTA |= (1<<PORTA2);
		}
		if (0x04 & c)
		{
			PORTA |= (1<<PORTA3);
		}
	}
	else if(i == 2){
		if (0x01 & c)
		{
			PORTA |= (1<<PORTA4);
		}
		if (0x02 & c)
		{
			PORTA |= (1<<PORTA5);
		}
		if (0x04 & c)
		{
			PORTA |= (1<<PORTA6);
		}
	}
	else if(i == 3){
		if (0x01 & c)
		{
			PORTA |= (1<<PORTA7);
		}
		if (0x02 & c)
		{
			PORTC |= (1<<PORTC7);
		}
		if (0x04 & c)
		{
			PORTC |= (1<<PORTC6);
		}
	}
	else if(i == 4){
		if (0x01 & c)
		{
			PORTC |= (1<<PORTC5);
		}
		if (0x02 & c)
		{
			PORTC |= (1<<PORTC4);
		}
		if (0x04 & c)
		{
			PORTC |= (1<<PORTC3);
		}
	}
	else if(i == 5){
		if (0x01 & c)
		{
			PORTC |= (1<<PORTC2);
		}
		if (0x02 & c)
		{
			PORTC |= (1<<PORTC1);
		}
		if (0x04 & c)
		{
			PORTC |= (1<<PORTC0);
		}
	}
	else if(i == 6){
		if (0x01 & c)
		{
			PORTD |= (1<<PORTD7);
		}
		if (0x02 & c)
		{
			PORTD |= (1<<PORTD6);
		}
		if (0x04 & c)
		{
			PORTD |= (1<<PORTD5);
		}
	}
	else if(i == 7){
		if (0x01 & c)
		{
			PORTD |= (1<<PORTD4);
		}
		if (0x02 & c)
		{
			PORTD |= (1<<PORTD3);
		}
		if (0x04 & c)
		{
			PORTD |= (1<<PORTD2);
		}
	}
}


// Turn off all the LEDs
void turn_off_all_LEDs()
{
	PORTA = 0;
	PORTB = 0x10;
	PORTC = 0;
	PORTD &= 0x03;
}


// Group the coefficients produced by the DCT to
// light up a set of LEDs depending on the
// sum of their energies
void DCT_light()
{	
	int low_thresh;
	int mid_thresh;
	int high_thresh;
	// Bitmasks used to indicate which LEDs to turn on
	char low_bitmask = 0x01;
	char mid_bitmask = 0x03;
	char high_bitmask = 0x07;

	// Make all outputs absolute value
	for(int i=0; i<32; i++)
	{
		if(S32[i]<0) S32[i] = -S32[i];
	}

	int group[8];

	switch(state) {
		case (DISP_ALL):
			low_thresh = 256;
			mid_thresh = 512;
			high_thresh = 1024;

	
			// Group the frequencies
			group[0] = S32[0]+ S32[1] + S32[2] + S32[3] - mid_thresh;
			group[1] = S32[4] + S32[5] + S32[6] + S32[7] - mid_thresh;
			group[2] = S32[8] + S32[9] + S32[10] + S32[11] - low_thresh;
			group[3] = S32[12] + S32[13] + S32[14] + S32[15] - low_thresh;
			group[4] = S32[16] + S32[17] + S32[18] + S32[19];
			group[5] = S32[20] + S32[21] + S32[22] + S32[23];
			group[6] = S32[24] + S32[25] + S32[26] + S32[27];
			group[7] = S32[28] + S32[29] + S32[30] + S32[31];
			break;
			
		case (DISP_LOWER_16):
			low_thresh = 256;
			mid_thresh = 512;
			high_thresh = 1024;

			group[0] = S32[0] + S32[1] - high_thresh;
			group[1] = S32[2] + S32[3] - low_thresh;
			group[2] = S32[4] + S32[5] - low_thresh;
			group[3] = S32[6] + S32[7];
			group[4] = S32[8] + S32[9];
			group[5] = S32[10] + S32[11];
			group[6] = S32[12] + S32[13];
			group[7] = S32[14] + S32[15];
			break;

		case (DISP_UPPER_16):
			low_thresh = 128;
			mid_thresh = 256;
			high_thresh = 512;

			group[0] = S32[16] + S32[17];
			group[1] = S32[18] + S32[19];
			group[2] = S32[20] + S32[21];
			group[3] = S32[22] + S32[23];
			group[4] = S32[24] + S32[25];
			group[5] = S32[26] + S32[27];
			group[6] = S32[28] + S32[29];
			group[7] = S32[30] + S32[31];
			break;

		case (DISP_ODD):
			low_thresh = 256;
			mid_thresh = 512;
			high_thresh = 1024;

			group[0] = S32[1] + S32[3];
			group[1] = S32[5] + S32[7];
			group[2] = S32[9] + S32[11];
			group[3] = S32[13] + S32[15];
			group[4] = S32[17] + S32[19];
			group[5] = S32[21] + S32[23];
			group[6] = S32[25] + S32[27];
			group[7] = S32[29] + S32[31];
			break;
		case (DISP_EVEN):
			low_thresh = 256;
			mid_thresh = 512;
			high_thresh = 1024;

			group[0] = S32[0] + S32[2];
			group[1] = S32[4] + S32[6];
			group[2] = S32[8] + S32[10];
			group[3] = S32[12] + S32[14];
			group[4] = S32[16] + S32[18];
			group[5] = S32[20] + S32[22];
			group[6] = S32[24] + S32[26];
			group[7] = S32[28] + S32[30];
			break;
	}

	
	
	// Turn off all lights
	turn_off_all_LEDs();	

	for (int i=0; i<8; i++)
	{
		char final_bitmask = 0x00;
		
		// Check if the sum of the group frequencies
		// is greater than the high threshold
		if(group[i] >=high_thresh)
		{
			final_bitmask = high_bitmask;
		}
		// Check if the sum of the group frequencies
		// is greater than the mid threshold
		else if(group[i] >= mid_thresh)
		{
			final_bitmask = mid_bitmask;
		}
		// Check if the sum of the group frequencies
		// is greater than the low threshold
		else if(group[i] >= low_thresh)
		{
			final_bitmask = low_bitmask;
		}
		// Turn on the group's LEDs based on final bitmask;
		turn_on_LED(final_bitmask, i);
	}
}

void init()
{
  // Set up port directions	
  DDRA = 0xFE;
  DDRB = 0xEF;
  DDRC = 0xFF;
  DDRD = 0xFE;
	
  PORTB = 0x10;

  // Set up timer 0
  TCCR0A = (1<<WGM01) | (1<<WGM00);
  // Prescale timer 0 by /8 so that the overflow ISR has freq of 8 khz
  TCCR0B = (1<<CS01 );

  // Initialize DC
  // Use AVCC as reference voltage
  ADMUX = ( 1<<REFS0 | 1<<ADLAR);
  // Turn on ADC, start conversion
  // Note: ADC conversion occurs on timer0 overflow
  ADCSRA = (1<<ADEN | 1<<ADSC) + 7;
  
  buffer_i = 0;
  proc_flag = 0;
  debounce_time = t1;
  PushState = 1;
  print_flag = 0;
  turn_off_all_LEDs();

  // Serial setop for debugging using printf, etc.     
  // Init the UART -- uart_init() is in uart.c
  uart_init();
  stdout = stdin = stderr = &uart_str;
  fprintf(stdout,"Modes of Operation \n\r");
  fprintf(stdout,"ALL:   Normal operation, displays all DCT coefficients \n\r");
  fprintf(stdout,"UPPER: Displays upper 16 DCT coefficients \n\r");
  fprintf(stdout,"LOWER: Displays lower 16 DCT coefficients \n\r");
  fprintf(stdout,"EVEN:  Displays all even DCT coefficients \n\r");
  fprintf(stdout,"ODD:   Displays all odd DCT coefficients \n\n\r");   
  
  // Constants for 8 sample DCT [also used by 16 sample)
  m1 = float2fix(cos(4*pi/16));
  m2 = float2fix(cos(6*pi/16));
  m3 = float2fix(cos(2*pi/16) - cos(6*pi/16));
  m4 = float2fix(cos(2*pi/16) + cos(6*pi/16));

  // Constants for 16 bit DCT 
  c1 = float2fix(0.5*cos(1.0*pi/32.0));
  c3 = float2fix(0.5*cos(3.0*pi/32.0));
  c5 = float2fix(0.5*cos(5.0*pi/32.0));
  c7 = float2fix(0.5*cos(7.0*pi/32.0));
  c9 = float2fix(0.5*cos(9.0*pi/32.0));
  c11 = float2fix(0.5*cos(11.0*pi/32.0));
  c13 = float2fix(0.5*cos(13.0*pi/32.0));
  c15 = float2fix(0.5*cos(15.0*pi/32.0)); 
  c17 = float2fix(0.5*cos(17.0*pi/32.0));
  c19 = float2fix(0.5*cos(19.0*pi/32.0));
  c21= float2fix(0.5*cos(21.0*pi/32.0));
  c23= float2fix(0.5*cos(23.0*pi/32.0));
  c25= float2fix(0.5*cos(25.0*pi/32.0));
  c27 = float2fix(0.5*cos(27.0*pi/32.0));
  c29 = float2fix(0.5*cos(29.0*pi/32.0));
  c31 = float2fix(0.5*cos(31.0*pi/32.0)); 
}

//=========================================================

int main(void)
begin

  init();
  int i;
  sei();

  // Turn on timer 0 overflow ISR
  TIMSK0 = (1<<TOIE0);

  // Initialize states
  state = DISP_ALL;
  next_state = DISP_ALL;

  fprintf(stdout,"Current mode: ALL\n");

  while(1){
  	state = next_state;
	if(proc_flag){
		proc_flag = 0;
		arrayCopy(bufferIn, a32); // Copy in the new sample
		normalizeForDCT(); 		  // Prepare data for the transform
		DCT1D_32(); 			  // Perform the transform
		DCT_light(); 			  // Adjust lights based on transform results
	}
	if(print_flag)
	{
		print_flag = 0;
		switch(next_state)
		begin
			case(DISP_ALL):
				fprintf(stdout,"Current mode: ALL\n");
				break;
			case(DISP_LOWER_16):
				fprintf(stdout,"Current mode: LOWER\n");
				break;
			case(DISP_UPPER_16):
				fprintf(stdout,"Current mode: UPPER\n");
				break;
			case(DISP_ODD):
				fprintf(stdout,"Current mode: ODD\n");
				break;
			case(DISP_EVEN):
				fprintf(stdout,"Current mode: EVEN\n");
				break;
		end
	}
  }
end
